import os
import shutil
import yaml
from enum import Enum

from tal.util import local_file_path


class Config(Enum):
    MITSUBA2_TRANSIENT_NLOS_FOLDER = 'MITSUBA2_TRANSIENT_NLOS_FOLDER'


def get_home_path():
    return os.path.expanduser('~')


def get_config_filename():
    return os.path.expanduser('~/.tal.conf')


def _parse_config(lines):
    return dict(line.split('=') for line in lines)


def ask_for_config(param_name: Config, force_ask=True):
    """Most useful: ask for a specific key in the config,
    and if it does not exist ask the user for a value"""
    config_dict = read_config()
    param_name = param_name.value
    if force_ask or param_name not in config_dict or len(config_dict[param_name].strip()) == 0:
        print(f'{param_name} is not specified. Please write a value:')
        param_value = ''
        while len(param_value.strip()) == 0:
            param_value = input(f'{param_name}=')
        config_dict[param_name] = param_value
        write_config(config_dict)
    return config_dict[param_name]


def read_config() -> dict:
    config = get_config_filename()
    if not os.path.isfile(config):
        print('TAL configuration file not found. '
              f'Creating a new one in {config}...')
        try:
            shutil.copy(local_file_path('.tal.conf.example'), config)
        except Exception as exc:
            print(f'/!\\ Unknown error when creating config file: {exc}')

    with open(config, 'r') as f:
        return _parse_config(f.readlines())


def write_config(config_dict):
    config = get_config_filename()
    with open(config, 'w') as f:
        f.writelines(list(f'{key}={value}'
                          for key, value in config_dict.items()))


def read_yaml(filename: str):
    with open(filename, 'r') as f:
        try:
            return yaml.safe_load(f)
        except yaml.YAMLError as exc:
            raise AssertionError(
                f'An error ocurred when parsing file ({filename}):\n{exc}')


def write_yaml_string(data_dict: dict) -> str:
    return yaml.dump(data_dict)



from tal.io.capture_data import NLOSCaptureData
from typing import Union, List
from enum import Enum


class ByAxis(Enum):
    UNKNOWN = 0
    T = 1
    X = 2
    Y = 3
    Z = 4


_Data = Union[NLOSCaptureData, NLOSCaptureData.HType]
_DataList = Union[List[_Data], _Data]


def xy_grid(data: Union[NLOSCaptureData, NLOSCaptureData.HType],
            size_x: int = 8, size_y: int = 8,
            t_start: int = None, t_end: int = None, t_step: int = 1):
    from tal.plot.xy import plot_xy_grid
    return plot_xy_grid(data, size_x, size_y, t_start, t_end, t_step)


def xy_interactive(data: _Data,
                   cmap: str = 'hot'):
    from tal.plot.xy import ByAxis, plot_txy_interactive
    return plot_txy_interactive(data, cmap, ByAxis.T)


def tx_interactive(data: _Data,
                   cmap: str = 'hot'):
    from tal.plot.xy import ByAxis, plot_txy_interactive
    return plot_txy_interactive(data, cmap, ByAxis.Y)


def ty_interactive(data: _Data,
                   cmap: str = 'hot'):
    from tal.plot.xy import ByAxis, plot_txy_interactive
    return plot_txy_interactive(data, cmap, ByAxis.X)


def t_comparison(data_list: _DataList,
                 x: int = None, y: int = None,
                 t_start: int = None, t_end: int = None,
                 a_min: float = None, a_max: float = None,
                 labels: List[str] = None):
    from tal.plot.compare import plot_t_comparison
    return plot_t_comparison(data_list, x, y, t_start, t_end, a_min, a_max, labels)


def txy_interactive(data: _Data,
                    cmap: str = 'hot', by: ByAxis = ByAxis.T):
    from tal.plot.xy import plot_txy_interactive
    return plot_txy_interactive(data, cmap, by)


def zxy_interactive(data: _Data,
                    cmap: str = 'hot', by: ByAxis = ByAxis.Z):
    from tal.plot.xy import plot_zxy_interactive
    return plot_zxy_interactive(data, cmap, by)



from tal.io.capture_data import NLOSCaptureData, write_hdf5
from tal.io.format import FileFormat, convert_dict


def read_capture(filename: str, file_format: FileFormat = FileFormat.AUTODETECT) -> NLOSCaptureData:
    return NLOSCaptureData(filename, file_format)


def write_capture(capture_data: NLOSCaptureData, filename: str, file_format: FileFormat = FileFormat.HDF5_TAL):
    write_hdf5(convert_dict(capture_data.as_dict(),
               format_to=file_format), filename)



import numpy as np
import matplotlib.pyplot as plt
import tal
from tal.enums import HFormat
from tqdm import tqdm


def backproject_pf_single_frequency(H_0, delta_t, d, wavelength, progress=True):
    """
    H_0: (nt, ns)
    delta_t: scalar, in meters
    d: (ns, nv)
    wavelength: scalar, in meters
    """
    assert H_0.ndim == 2, 'Incorrect H format'

    ns, nv = d.shape
    nt, ns_ = H_0.shape
    assert ns == ns_, 'Incorrect shape'

    propagator = np.exp(2j * np.pi * d / wavelength)

    t = delta_t * np.linspace(start=0, stop=nt - 1, num=nt)
    e = np.exp(-2j * np.pi * t / wavelength)
    H_0_w = np.sum(H_0 * e.reshape((nt, 1)), axis=0).reshape((ns, 1))
    # FIXME(diego): implement convolution in frequency domain
    H_1_w = np.sum(H_0_w * propagator, axis=0)

    return H_1_w


def backproject_pf_multi_frequency(
        H_0, laser_grid_xyz, sensor_grid_xyz, volume_xyz_n3,
        camera_system, t_accounts_first_and_last_bounces,
        t_start, delta_t,
        wl_mean, wl_sigma, edges,
        laser_xyz=None, sensor_xyz=None, progress=False):
    assert H_0.ndim == 2, 'Incorrect H format'
    assert volume_xyz_n3.ndim == 2 and volume_xyz_n3.shape[1] == 3, \
        'Incorrect volume_xyz format, should be N_3'
    assert laser_grid_xyz.size == 3, 'Only supports one laser position'
    assert (not t_accounts_first_and_last_bounces or (laser_xyz is not None and sensor_xyz is not None)), \
        't_accounts_first_and_last_bounces requires laser_xyz and sensor_xyz'

    nt, ns = H_0.shape
    nv, _ = volume_xyz_n3.shape

    """ Phasor fields filter """

    # FIXME(diego): obtain filter parameters directly in frequency
    # domain (avoid conversion from time domain)

    t_6sigma = int(np.ceil(6 * wl_sigma / delta_t))
    padding = 2 * t_6sigma
    # FIXME(diego) if we want to convert a circular convolution to linear,
    # this should be nt + t_6sigma - 1 instead of nt + 4 * t_6sigma or even nt + 2 * t_6sigma
    # I have found cases where it fails even with nt + 2 * t_6sigma (Z, 0th pixel)
    nf = nt + 2 * padding

    t_max = delta_t * (nf - 1)
    t = np.linspace(start=0, stop=t_max, num=nf)

    gaussian_envelope = np.exp(-((t - t_max / 2) / wl_sigma) ** 2 / 2)
    K = gaussian_envelope / np.sum(gaussian_envelope) * \
        np.exp(2j * np.pi * t / wl_mean)
    K = np.fft.fftshift(K)  # center at zero

    mean_idx = (nf * delta_t) / wl_mean
    sigma_idx = (nf * delta_t) / (wl_sigma * 6)
    # shift to center at zero, easier for low negative frequencies
    freq_min_idx = nf // 2 + int(np.floor(mean_idx - 3 * sigma_idx))
    freq_max_idx = nf // 2 + int(np.ceil(mean_idx + 3 * sigma_idx))
    K_fftshift = np.fft.fftshift(np.fft.fft(K))
    K_fftfreq = np.fft.fftshift(np.fft.fftfreq(nf, d=delta_t))

    weights = K_fftshift[freq_min_idx:freq_max_idx+1]
    freqs = K_fftfreq[freq_min_idx:freq_max_idx+1]
    nw = len(weights)

    if edges == 'zero':
        H_0 = np.pad(H_0, ((padding, padding), (0, 0)), 'constant')
    else:
        raise AssertionError('Implemented only for edges="zero"')

    """ Propagation of specific frequencies """

    # d_3
    d = np.linalg.norm(
        sensor_grid_xyz.reshape((ns, 1, 3)) -
        volume_xyz_n3.reshape((1, nv, 3)),
        axis=2)
    d -= t_start
    if t_accounts_first_and_last_bounces:
        # d_1
        d += np.linalg.norm(laser_xyz.reshape(3) -
                            laser_grid_xyz.reshape(3))
        # d_4
        d += np.linalg.norm(sensor_xyz.reshape((1, 1, 3)) -
                            sensor_grid_xyz.reshape((ns, 1, 3)), axis=2)
    if camera_system.bp_accounts_for_d_2():
        # d_2
        d += np.linalg.norm(
            laser_grid_xyz.reshape((1, 1, 3)) -
            volume_xyz_n3.reshape((1, nv, 3)),
            axis=2)

    if camera_system.is_transient():
        H_1 = np.zeros((nt, nv), dtype=np.complex64)
    else:
        H_1 = np.zeros(nv, dtype=np.complex64)

    t = delta_t * np.linspace(start=0, stop=nf - 1, num=nf)

    iterator = zip(freqs, weights)
    if progress:
        iterator = tqdm(iterator, total=nw)

    for frequency, weight in iterator:
        wavelength = np.inf if np.isclose(frequency, 0) else 1 / frequency
        H_1_w = weight * \
            backproject_pf_single_frequency(
                H_0, delta_t, d, wavelength, progress=progress)
        e = np.exp(2 * np.pi * 1j * t * frequency) / nf
        H_1_i = H_1_w.reshape((1, nv)) * e.reshape((nf, 1))

        if camera_system.is_transient():
            H_1 += H_1_i[padding: -padding, ...]
        else:
            H_1 += H_1_i[padding]

    return H_1



import tal
from tal.plot import ByAxis
print('Start read')
data = tal.io.read_capture('test/data/ZNLOS_Z_10_Single.hdf5')
print('Finish read')


tal.plot.xy_grid(data, size_x=8, size_y=4)

tal.plot.txy_interactive(data, by = ByAxis.T)
tal.plot.txy_interactive(data, by = ByAxis.X)
tal.plot.txy_interactive(data, by = ByAxis.Y)

tal.plot.zxy_interactive(data.H, by = ByAxis.Z)
tal.plot.zxy_interactive(data.H, by = ByAxis.X)
tal.plot.zxy_interactive(data.H, by = ByAxis.Y)





"""
Author      :   Pablo Luesia Lahoz
E-Mail      :   pluesia at unizar.es
File        :   rsd_kernel.py
Description :   Class for a RSD kernel generator, and propagator, in order
                to perform fast propagation from plane to parallel plane
"""

import numpy as np


class RSD_kernel(object):
    """
    Class implementation for a RSD kernel. It creates a kernel for performance
    fast propagation between paralel planes. 
    """

    def __init__(self, V: np.ndarray, P: np.ndarray, lambda_v: np.ndarray):
        """
        Initialize the class of the kernel. The plane P and the planes in V has
        to be parallel.
        @param V            : Volume in 3d to propagate, formed by parallel
                              planes. The first axis are the planes, the second
                              are the rows, the third the columns, and the 
                              forth the coordinates of the points in the volume.
        @param P            : Points of the plane origin to propagate. The
                              first axis are the rows, the second the columns,
                              and the third the coordinates of the points in 
                              the plane.
        @param lambda_v     : Arrays of wavelengths to performance the RSD
                              propagation in metters.
        :return             : An instance of RSD_kernel class.
        """
        # Check the planes shapes are the same
        assert P.ndim == 3 and V.ndim == 4, \
            "P has to be a plane, and V a volume. Both with 3d coordinates"
        assert P.shape == V[0].shape, "V and P shapes are not the same"
        # Check the planes are parallel
        # First V plane normal
        normal_V, d_V = RSD_kernel.__as_plane_eq(V[0, 0, 0],
                                                 V[0, -1, 0],
                                                 V[0, -1, -1])
        normal_P, d_P = RSD_kernel.__as_plane_eq(P[0, 0], P[-1, 0], P[-1, -1])
        # Check if planes are parallel (with an small error to 0)
        assert np.linalg.norm(np.cross(normal_V, normal_P)) == 0, \
            "V and P are not parallel"

        # Volume to propagate
        self._V = V
        self._n_V = normal_V
        self._d_V = d_V
        # Origin plane to propagate
        self._P = P
        self._n_P = normal_P
        self._d_P = d_P
        # Wavelengths to propagate
        self._wl_v = lambda_v
        # Preprocess values. It is None until data is preprocessed
        self._preprocess = (self.__K_at_P(), self.__vecs_P_to_V_planes())

    ###########################################################################
    #                             PRIVATE METHODS                             #
    ###########################################################################

    # Given 3 points it return the plane equation

    def __as_plane_eq(p1, p2, p3):
        # Vectors in plane
        v1 = p3 - p1
        v2 = p2 - p1
        # Normal to the plane
        n = np.cross(v1, v2)
        d = np.dot(n, p3)
        return n, d

    # Generate a coordinates matrix kernel for the plane P (at distance 0)

    def __K_at_P(self):
        # Precalculate kernel vectors
        (nx, ny, _) = self._P.shape
        # Kernel shape
        rx = 2*nx - 1
        ry = 2*ny - 1
        # Delta of the points in i and j axis of the rectangle
        di = self._P[0, 1] - self._P[0, 0]
        dj = self._P[1, 0] - self._P[0, 0]
        # Generate the central row, and from it the matrix
        K_row = np.linspace(start=-di * (nx - 1),
                            stop=di * (nx - 1),
                            num=rx, dtype=np.float64)
        K_base = np.linspace(start=K_row - dj * (ny - 1),
                             stop=K_row + dj * (ny - 1),
                             num=ry, dtype=np.float64)
        return K_base

    # Return an array of vectors from P to the different V planes

    def __vecs_P_to_V_planes(self):
        n_V_m = np.linalg.norm(self._n_V)   # Modulus of V first plane normal
        # Distance from P to V first plane
        d_P2V = np.abs(np.sum(self._n_V * self._P[0, 0]) + self._d_V) / n_V_m
        # Unitary normal vector of V planes
        n_V_i = self._n_V / n_V_m
        # number of V planes
        rz = self._V.shape[0]
        # Distance between V planes
        d_V2V = np.linalg.norm(self._V[0, 0, 0] - self._V[-1, 0, 0]) / (rz - 1)
        # Shortest vector from first V plane to the V planes
        n_V2V = np.linspace(start=n_V_i*0,            # Vector to itself
                            stop=n_V_i * d_V2V * rz,  # Vector furthest plane
                            num=rz, dtype=np.float64)
        # Add the vector from P to V
        return n_V2V + n_V_i*d_P2V

    # Return a spatial kernel with the vecotrs from P to the i-th plane in V.
    def __spatial_i_kernel(self, i):
        K_base, v_P2V = self._preprocess
        return K_base + v_P2V[i]

    ###########################################################################
    #                             PUBLIC METHODS                              #
    ###########################################################################
    def kernel_shape(self):
        """
        Returns the shape of the Kernel
        """
        (K_base, _) = self._preprocess
        return (K_base.shape[0], K_base.shape[1])

    def get_f_RSD_kernel_i(self, i: int):
        """
        Returns the RSD kernel in the 2d fourier domain of the 
        @param i    : Index of the V plane
        :return     : The plane kernel for i-th plane of V in fourier domain
        """
        s_K = self.__spatial_i_kernel(i)
        d_K = np.linalg.norm(s_K, axis=-1)
        rsd_K = RSD_kernel.RSD_kernel_w(d_K, self._wl_v)
        return np.fft.fft2(rsd_K)

    @staticmethod
    def RSD_kernel_w(K_d: np.ndarray, lambda_w: float):
        """
        Given a rectangular distance spatial kernel K_d, it returns the 
        Rayleigh-Sommerfled propagator kernel for the frequency with wavelength
        lambda_w.
        @param K_d      : Rectangular distance spatial kernel
        @param lambda_w : Wavelength to performance the transformation
        @return         : Complex matrix representing the kernel
        """
        # Broadcastable lambda_w to match K_d
        lambda_w_br = lambda_w.reshape((-1,) + (1,) * K_d.ndim)
        return np.exp(2*np.pi*1j/lambda_w_br * K_d)/K_d



from nptyping import NDArray
from tal.io.capture_data import NLOSCaptureData
from tal.enums import HFormat
from tal.plot import ByAxis
from tal.util import SPEED_OF_LIGHT
from typing import Union
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
from matplotlib.widgets import Slider, Button
import numpy as np
from tqdm import tqdm


def plot_xy_grid(data: Union[NLOSCaptureData, NLOSCaptureData.HType],
                 size_x: int, size_y: int,
                 t_start: int, t_end: int, t_step: int):
    def t_to_time(t): return f'Bin #{t_start + t * t_step}'
    if isinstance(data, NLOSCaptureData):
        assert data.H_format == HFormat.T_Sx_Sy, \
            'plot_xy_grid does not support this data format'
        txy = data.H
        if data.t_start is not None and data.delta_t is not None:
            def t_to_time(
                t): return f'Bin #{(t_start or 0) + t * t_step}, {(data.t_start + t * data.delta_t) * 1e12 / SPEED_OF_LIGHT:.0f} ps'
    else:
        assert data.ndim == 3 and data.shape[1] == data.shape[2], \
            'plot_xy_grid does not support this data format'
        txy = data
    txy = txy[t_start:t_end:t_step, ...]
    nt = txy.shape[0]
    step = 1
    plot_size = size_x * size_y
    while nt // step > plot_size:
        step *= 2
    txy_min, txy_max = np.min(txy), np.max(txy)
    fig, axs = plt.subplots(size_y, size_x)

    for i in tqdm(range(plot_size)):
        t_bin = i * step
        image = txy[t_bin]
        row = i // size_x
        col = i % size_x
        mappable = axs[row, col].imshow(image.astype(
            np.float32), cmap='jet', vmin=txy_min, vmax=txy_max)
        fig.colorbar(mappable, ax=axs[row, col])
        axs[row, col].axis('off')
        axs[row, col].set_title(t_to_time((t_start or 0) + t_bin * t_step))

    plt.tight_layout()
    plt.show()


def plot_3d_interactive_axis(xyz: np.ndarray, focus_slider: np.ndarray,
                             axis: int, title: str, slider_title: str,
                             slider_unit: str, cmap: str = 'hot',
                             xlabel: str = '', ylabel: str = ''):
    """More general plotting"""
    assert xyz.ndim == 3, 'Unknown H_Format to plot'
    assert axis < 3, f'Data only has 3 dims (given axis={axis})'
    assert xyz.shape[axis] == len(focus_slider), \
        'The slider and the data have different lengths'
    # Move the axis, so the interactive axis is at 0
    xyz_p = np.moveaxis(xyz, axis, 0)
    max_idx = xyz_p.shape[0] - 1
    v_min = np.min(xyz)
    v_max = np.max(xyz)

    # Plot the first figure
    fig = plt.figure()
    ax = plt.axes([0.2, 0.2, 0.7, 0.7])
    cax = plt.axes([0.85, 0.2, 0.03, 0.7])
    colorbar = None
    scale = 'linear'

    def set_ax_common():
        nonlocal colorbar
        fig.suptitle(title)
        if colorbar is not None:
            cax.cla()
        colorbar = fig.colorbar(img, cax=cax, shrink=0.5)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.set_xticks([])
        ax.set_yticks([])

    def scale_linear(event):
        nonlocal img, scale
        scale = 'linear'
        ax.cla()
        img = ax.imshow(xyz_p[current_idx], cmap=cmap, vmin=v_min, vmax=v_max)
        set_ax_common()

    def scale_log(event):
        nonlocal img, scale
        scale = 'log'
        ax.cla()
        img = ax.imshow(xyz_p[current_idx] + 1, cmap=cmap,
                        norm=LogNorm(vmin=v_min + 1, vmax=v_max))
        set_ax_common()

    current_idx = 0
    img = None
    scale_linear(None)

    ax_slider = plt.axes([0.25, 0.1, 0.5, 0.03])
    slider = Slider(
        ax=ax_slider,
        label=slider_title,
        valmin=0,
        valmax=len(focus_slider) - 1,
        valinit=0,
        valstep=1.0,
        orientation='horizontal')
    ax_text = plt.axes([0.25, 0.05, 0.5, 0.03])
    ax_text.axis('off')
    text = ax_text.text(0, 0, f'{focus_slider[0]:.3f} {slider_unit}')

    ax_button_linear = plt.axes([0.05, 0.4, 0.1, 0.05])
    button_linear = Button(ax_button_linear, 'Linear')
    button_linear.on_clicked(scale_linear)
    ax_button_log = plt.axes([0.05, 0.45, 0.1, 0.05])
    button_log = Button(ax_button_log, 'Log')
    button_log.on_clicked(scale_log)

    def cmap_hot(event):
        nonlocal cmap
        cmap = 'hot'
        if scale == 'linear':
            scale_linear(None)
        else:
            scale_log(None)

    def cmap_nipyspectral(event):
        nonlocal cmap
        cmap = 'nipy_spectral'
        if scale == 'linear':
            scale_linear(None)
        else:
            scale_log(None)

    ax_button_cmap_hot = plt.axes([0.05, 0.55, 0.1, 0.05])
    button_cmap_hot = Button(ax_button_cmap_hot, 'hot')
    button_cmap_hot.on_clicked(cmap_hot)
    ax_button_nipyspectral = plt.axes([0.05, 0.6, 0.1, 0.05])
    button_nipyspectral = Button(
        ax_button_nipyspectral, 'nipy_spectral')
    button_nipyspectral.on_clicked(cmap_nipyspectral)

    # Figure update
    def update(i):
        nonlocal current_idx
        current_idx = int(i)
        img.set_array(xyz_p[current_idx] + (1 if scale == 'log' else 0))
        text.set_text(f'{focus_slider[current_idx]:.3f} {slider_unit}')

    def update_prev(event):
        slider.set_val(0 if current_idx == 0 else current_idx - 1)

    def update_next(event):
        slider.set_val(max_idx if current_idx == max_idx else current_idx + 1)

    ax_prev = plt.axes([0.15, 0.1, 0.03, 0.03])
    button_prev = Button(ax_prev, '<')
    button_prev.on_clicked(update_prev)
    ax_next = plt.axes([0.8, 0.1, 0.03, 0.03])
    button_next = Button(ax_next, '>')
    button_next.on_clicked(update_next)

    slider.on_changed(update)
    plt.show()


def plot_txy_interactive(data: Union[NLOSCaptureData, NLOSCaptureData.HType],
                         cmap: str = 'hot', by: ByAxis = ByAxis.T):
    if isinstance(data, NLOSCaptureData):
        assert data.H_format == HFormat.T_Sx_Sy, \
            'plot_txy_interactive does not support this data format'
        txy = data.H
        delta_t = data.delta_t
    else:
        assert data.ndim == 3, \
            'plot_txy_interactive does not support this data format'
        txy = data
        delta_t = None

    title = 'Impulse response '
    # Plot parameters
    if by == ByAxis.T:
        axis = 0
        n_it = txy.shape[axis]
        it_v = np.arange(n_it, dtype=np.float32)
        title += 'by time'
        slider_title = 'Bins'
        slider_unit = 'index'
        if delta_t is not None:
            it_v *= delta_t
            slider_unit = 'ps'
        xlabel = 'x'
        ylabel = 'y'
    elif by == ByAxis.Y:
        axis = 1
        n_it = txy.shape[axis]
        it_v = np.arange(n_it, dtype=np.float32)
        title += 'by y'
        slider_title = 'Planes'
        slider_unit = 'index'
        txy = txy.swapaxes(0, 2)
        xlabel = 't'
        ylabel = 'x'
    elif by == ByAxis.X:
        axis = 2
        n_it = txy.shape[axis]
        it_v = np.arange(n_it, dtype=np.float32)
        title += 'by x'
        slider_title = 'Planes'
        slider_unit = 'index'
        txy = txy.swapaxes(0, 1)
        xlabel = 't'
        ylabel = 'y'
    else:
        raise 'plot_txy_interactive does not support this axis'

    # Plot the data
    return plot_3d_interactive_axis(txy, it_v, axis=axis,
                                    title=title,
                                    slider_title=slider_title,
                                    slider_unit=slider_unit,
                                    cmap=cmap,
                                    xlabel=xlabel, ylabel=ylabel)


def plot_zxy_interactive(data: NDArray, cmap: str = 'hot', by: ByAxis = ByAxis.Z):

    assert data.ndim == 3, \
        'plot_zxy_interactive does not support this data format'
    zxy = data

    title = 'Amplitude by '
    # Plot parameters
    if by == ByAxis.Z:
        axis = 0
        n_it = zxy.shape[axis]
        it_v = np.arange(n_it, dtype=np.float32)
        title += 'by z'
        slider_title = 'Planes'
        slider_unit = 'index'
        xlabel = 'x'
        ylabel = 'y'
    elif by == ByAxis.Y:
        axis = 1
        n_it = zxy.shape[axis]
        it_v = np.arange(n_it, dtype=np.float32)
        title += 'by y'
        slider_title = 'Planes'
        slider_unit = 'index'
        xlabel = 'x'
        ylabel = 'z'
    elif by == ByAxis.X:
        axis = 2
        n_it = zxy.shape[axis]
        it_v = np.arange(n_it, dtype=np.float32)
        title += 'by x'
        slider_title = 'Planes'
        slider_unit = 'index'
        xlabel = 'y'
        ylabel = 'z'
    else:
        raise 'plot_zxy_interactive does not support this axis'

    # Plot the data
    return plot_3d_interactive_axis(zxy, it_v, axis=axis,
                                    title=title,
                                    slider_title=slider_title,
                                    slider_unit=slider_unit,
                                    cmap=cmap,
                                    xlabel=xlabel, ylabel=ylabel)



from tal.enums import HFormat, GridFormat
import numpy as np
from tqdm import tqdm


def backproject(H_0, laser_grid_xyz, sensor_grid_xyz, volume_xyz,
                camera_system, t_accounts_first_and_last_bounces,
                t_start, delta_t,
                laser_xyz=None, sensor_xyz=None, progress=False):
    # TODO(diego): extend for multiple laser points
    assert H_0.ndim == 2 and laser_grid_xyz.size == 3, \
        'backproject only supports one laser point'

    nt, ns = H_0.shape
    assert sensor_grid_xyz.shape[0] == ns, 'H does not match with sensor_grid_xyz'
    assert (not t_accounts_first_and_last_bounces or (laser_xyz is not None and sensor_xyz is not None)), \
        't_accounts_first_and_last_bounces requires laser_xyz and sensor_xyz'
    ns, _ = sensor_grid_xyz.shape
    nv, _ = volume_xyz.shape

    if camera_system.is_transient():
        H_1 = np.zeros((nt, nv), dtype=H_0.dtype)
    else:
        H_1 = np.zeros(nv, dtype=H_0.dtype)

    # d_1: laser origin to laser illuminated point
    # d_2: laser illuminated point to x_v
    # d_3: x_v to sensor imaged point
    # d_4: sensor imaged point to sensor origin
    x_l = laser_grid_xyz.reshape(3)
    if t_accounts_first_and_last_bounces:
        d_1 = np.linalg.norm(laser_xyz - x_l)
    else:
        d_1 = 0.0

    range_s = range(ns)
    if progress:
        range_s = tqdm(range_s)
    for s_i in range_s:
        x_s = sensor_grid_xyz[s_i, :]
        if t_accounts_first_and_last_bounces:
            d_4 = np.linalg.norm(x_s - sensor_xyz)
        else:
            d_4 = 0.0
        for i_v, x_v in enumerate(volume_xyz):
            if camera_system.bp_accounts_for_d_2():
                d_2 = np.linalg.norm(x_l - x_v)
            else:
                d_2 = 0.0
            d_3 = np.linalg.norm(x_v - x_s)
            t_i = int((d_1 + d_2 + d_3 + d_4 - t_start) / delta_t)
            if camera_system.is_transient():
                p = np.copy(H_0[:, s_i])
                if t_i > 0:
                    p[:t_i] = 0.0
                elif t_i < 0:
                    p[t_i+nt-1:] = 0.0
                H_1[:, i_v] += np.roll(p, -t_i)
            else:
                if t_i >= 0 and t_i < nt:
                    H_1[i_v] += H_0[t_i, s_i]

    return H_1



from tal.reconstruct.bp import *
from tal.reconstruct.pf import *
from tal.reconstruct.pf_dev import *
from tal.io.capture_data import NLOSCaptureData
from tal.enums import HFormat
from typing import Union

_Data = Union[NLOSCaptureData, NLOSCaptureData.HType]


def filter_H(data: _Data,
             filter_name: str,
             data_format: HFormat = HFormat.UNKNOWN,
             border: str = 'zero',
             plot_filter: bool = False,
             return_filter: bool = False,
             **kwargs) -> NLOSCaptureData.HType:
    """
    Filter a captured data signal (H) using specified filter_name
        * data_format should be non-null if not specified through data
        * border sets the behaviour for the edges of the convolution
          - 'erase': the filtered signal has the edges set to zero
          - 'zero': before filtering, pad the signal with zeros
          - 'edge': before filtering, pad the signal with edge values
        * If plot_filter=True, shows a plot of the resulting filter
        * If return_filter=True, returns the filter (K)
          else, returns the filtered signal (H * K)
    Available filters and respective arguments:
    - pf: Filter certain frequencies, weighted using a Gaussian on the frequency domain
        * wl_mean: Mean of the Gaussian in the frequency domain
        * wl_sigma: STD of the Gaussian in the frequency domain
        * delta_t: Time interval, must be non-null if not specified through data
        FIXME(diego): is this sentence true? vvv probably not
        e.g. mean = 3, sigma = 0.5 will filter frequencies of ~2-4m
    """
    from tal.reconstruct.filters import filter_H_impl
    return filter_H_impl(data, filter_name, data_format, border, plot_filter, return_filter, **kwargs)


def get_volume_min_max_resolution(minimal_pos, maximal_pos, resolution):
    assert np.all(maximal_pos > minimal_pos)

    e = resolution / 2  # half-voxel
    x = np.arange(minimal_pos[0] + e, maximal_pos[0], resolution)
    y = np.arange(minimal_pos[1] + e, maximal_pos[1], resolution)
    z = np.arange(minimal_pos[2] + e, maximal_pos[2], resolution)
    nx, ny, nz = len(x), len(y), len(z)
    x = np.repeat(x.reshape(nx, 1, 1), ny, axis=1)
    x = np.repeat(x.reshape(nx, ny, 1), nz, axis=2)
    y = np.repeat(y.reshape(1, ny, 1), nx, axis=0)
    y = np.repeat(y.reshape(nx, ny, 1), nz, axis=2)
    z = np.repeat(z.reshape(1, 1, nz), nx, axis=0)
    z = np.repeat(z.reshape(nx, 1, nz), ny, axis=1)

    return np.stack((x, y, z), axis=-1)



from tal.io.capture_data import NLOSCaptureData
import matplotlib.pyplot as plt
import numpy as np
import tal
from matplotlib.widgets import Slider, RangeSlider, Button
from matplotlib.colors import LogNorm
from tal.enums import HFormat


def plot_t_comparison(data_list, x, y, t_start, t_end, a_min, a_max, labels):
    if isinstance(data_list, np.ndarray) or isinstance(data_list, NLOSCaptureData):
        data_list = [data_list]
    assert all((isinstance(data, np.ndarray) and data.ndim == 3) or
               (isinstance(data, NLOSCaptureData) and data.H_format == HFormat.T_Sx_Sy) for data in data_list), \
        'Incorrect data types or HFormat, this function only works for HFormat = T_Sx_Sy'

    data_real = list(map(
        lambda x:
        np.real(x.H) if isinstance(x, NLOSCaptureData)
        else np.real(x),
        data_list))

    def get_H(data):
        if isinstance(data, NLOSCaptureData):
            return data.H
        else:
            return data

    nt, ny, nx = get_H(data_real[0]).shape
    for data in data_real[1:]:
        nt2, ny2, nx2 = get_H(data).shape
        assert nt == nt2 and ny == ny2 and nx == nx2, \
            'Dimensions do not match'

    fig = plt.figure()
    fig.suptitle('Signal comparison')

    ax = plt.axes([0.2, 0.2, 0.7, 0.7])
    x_slider_ax = plt.axes([0.25, 0.13, 0.5, 0.03])
    y_slider_ax = plt.axes([0.25, 0.10, 0.5, 0.03])
    t_slider_ax = plt.axes([0.25, 0.07, 0.5, 0.03])
    a_slider_ax = plt.axes([0.25, 0.04, 0.5, 0.03])

    x = x or 0
    y = y or 0
    t_start = t_start or 0
    t_end = t_end or nt - 1
    A_MIN_TOTAL = min(np.min(get_H(data)) for data in data_real)
    A_MAX_TOTAL = max(np.max(get_H(data)) for data in data_real)
    a_min = a_min or A_MIN_TOTAL
    a_max = a_max or A_MAX_TOTAL

    def get_q001(H):
        if np.any(H > 0):
            return np.quantile(H[H > 0], 0.001)
        else:
            return 0.0

    q001 = np.min(list(get_q001(get_H(data)) for data in data_real))
    scale = 'linear'

    def update():
        ax.cla()
        x_range = list(range(t_start, t_end, 1))
        for i, data in enumerate(data_real):
            if scale == 'linear':
                ax.plot(x_range, get_H(data)[t_start:t_end, x, y],
                        label=str(i) if labels is None else labels[i])
                ax.set_ylim(bottom=a_min, top=a_max)
            else:
                ax.plot(x_range, get_H(data)[t_start:t_end, x, y] + q001,
                        label=str(i) if labels is None else labels[i])
                ax.set_ylim(bottom=a_min + q001, top=a_max)
            ax.legend()
        ax.set_yscale(scale)
        plt.draw()

    x_slider = Slider(
        ax=x_slider_ax, label='x', valmin=0, valmax=nx - 1,
        valinit=x, valstep=1, orientation='horizontal')
    y_slider = Slider(
        ax=y_slider_ax, label='y', valmin=0, valmax=ny - 1,
        valinit=y, valstep=1, orientation='horizontal')
    t_slider = RangeSlider(
        ax=t_slider_ax, label='t', valmin=0, valmax=nt - 1,
        valinit=(t_start, t_end), valstep=1, orientation='horizontal')
    a_slider = RangeSlider(
        ax=a_slider_ax, label='Amplitude', valmin=A_MIN_TOTAL, valmax=A_MAX_TOTAL,
        valinit=(a_min, a_max), orientation='horizontal')

    def update_x(new_x):
        nonlocal x
        x = new_x
        update()

    def update_y(new_y):
        nonlocal y
        y = new_y
        update()

    def update_t(new_t):
        nonlocal t_start, t_end
        t_start, t_end = new_t
        update()

    def update_a(new_a):
        nonlocal a_min, a_max
        a_min, a_max = new_a
        update()

    x_slider.on_changed(update_x)
    y_slider.on_changed(update_y)
    t_slider.on_changed(update_t)
    a_slider.on_changed(update_a)

    def scale_linear(_):
        nonlocal scale
        scale = 'linear'
        update()

    def scale_log(_):
        nonlocal scale
        scale = 'log'
        update()

    ax_button_linear = plt.axes([0.03, 0.4, 0.1, 0.05])
    button_linear = Button(ax_button_linear, 'Linear')
    button_linear.on_clicked(scale_linear)
    ax_button_log = plt.axes([0.03, 0.45, 0.1, 0.05])
    button_log = Button(ax_button_log, 'Log')
    button_log.on_clicked(scale_log)

    update()
    plt.show()



from tal.enums import HFormat, GridFormat, VolumeFormat, CameraSystem
from tal.io.capture_data import NLOSCaptureData
import numpy as np


def convert_to_N_3(data: NLOSCaptureData,
                   volume_xyz: NLOSCaptureData.VolumeXYZType,
                   volume_format: VolumeFormat):
    if data.H_format == HFormat.T_Si:
        H = data.H
    elif data.H_format == HFormat.T_Sx_Sy:
        nt, nsx, nsy = data.H.shape
        H = data.H.reshape(nt, nsx * nsy)
    else:
        raise AssertionError(f'H_format {data.H_format} not implemented')

    # FIXME(diego) also convert laser data and confocal/exhaustive H measurements

    if data.sensor_grid_format == GridFormat.N_3:
        sensor_grid_xyz = data.sensor_grid_xyz
    elif data.sensor_grid_format == GridFormat.X_Y_3:
        try:
            assert nsx == data.sensor_grid_xyz.shape[0] and nsy == data.sensor_grid_xyz.shape[1], \
                'sensor_grid_xyz.shape does not match with H.shape'
        except NameError:
            # nsx, nsy not defined, OK
            nsx, nsy, _ = data.sensor_grid_xyz.shape
            pass
        sensor_grid_xyz = data.sensor_grid_xyz.reshape(nsx * nsy, 3)
    else:
        raise AssertionError(
            f'sensor_grid_format {data.sensor_grid_format} not implemented')

    assert H.shape[1] == sensor_grid_xyz.shape[0], \
        'H.shape does not match with sensor_grid_xyz.shape. Different number of points than measurements.'

    # FIXME add X_Y_3 case (check other VolumeFormats maybe)
    if volume_format == VolumeFormat.X_Y_Z_3:
        volume_xyz_n3 = volume_xyz.reshape((-1, 3))
    elif volume_format == VolumeFormat.N_3:
        volume_xyz_n3 = volume_xyz
    else:
        raise AssertionError('volume_format must be specified')

    return (H, data.laser_grid_xyz, sensor_grid_xyz, volume_xyz_n3)


def convert_reconstruction_from_N_3(data: NLOSCaptureData,
                                    reconstructed_volume_n3: np.ndarray,  # FIXME type
                                    volume_xyz: NLOSCaptureData.VolumeXYZType,
                                    volume_format: VolumeFormat,
                                    camera_system: CameraSystem):
    if camera_system.is_transient():
        time_dim = (data.H.shape[data.H_format.time_dim()],)
    else:
        time_dim = ()

    assert volume_format.xyz_dim_is_last(), 'Unexpected volume_format'
    return reconstructed_volume_n3.reshape(
        time_dim + volume_xyz.shape[:-1])



from tal.io.capture_data import NLOSCaptureData
from tal.enums import VolumeFormat, CameraSystem
from tal.reconstruct.utils import convert_to_N_3, convert_reconstruction_from_N_3
import numpy as np


def solve(data: NLOSCaptureData,
          wl_mean: float,
          wl_sigma: float,
          edges: str = 'zero',  # FIXME: rename to border
          volume_xyz: NLOSCaptureData.VolumeXYZType = None,
          volume_format: VolumeFormat = None,
          camera_system: CameraSystem = CameraSystem.STEADY) -> np.array:  # FIXME type
    """
    NOTE: Does _NOT_ attempt to compensate effects caused by attenuation:
      - cos decay i.e. {sensor|laser}_grid_normals are ignored
      - 1/d^2 decay
    TODO(diego): docs
    """
    H, laser_grid_xyz, sensor_grid_xyz, volume_xyz_n3 = \
        convert_to_N_3(data, volume_xyz, volume_format)

    from tal.reconstruct.pf_dev.phasor_fields import backproject_pf_multi_frequency
    reconstructed_volume_n3 = backproject_pf_multi_frequency(
        H, laser_grid_xyz, sensor_grid_xyz, volume_xyz_n3,
        camera_system, data.t_accounts_first_and_last_bounces,
        data.t_start, data.delta_t,
        wl_mean, wl_sigma, edges,
        data.laser_xyz, data.sensor_xyz,
        progress=True)

    return convert_reconstruction_from_N_3(data, reconstructed_volume_n3, volume_xyz, volume_format, camera_system)



import os
import shutil
import yaml
import tal
from tal import __version__ as tal_version
from tal.io.capture_data import NLOSCaptureData
from tal.enums import FileFormat, GridFormat, HFormat
from tal.util import fdent, write_img, tonemap_ldr
from tal.render.mitsuba2_transient_nlos import (
    get_material_keys, get_materials,
    read_mitsuba_streakbitmap, read_mitsuba_bitmap,
    run_mitsuba, mitsuba_set_variant
)
from tal.config import local_file_path
import datetime
import numpy as np
from tqdm import tqdm


def get_scene_xml(config, random_seed=0, quiet=False):
    def s(value):
        if isinstance(value, bool):
            return str(value).lower()
        else:
            return value

    def v(key):
        return s(config[key])

    # integrator
    integrator_steady = fdent(f'''\
        <integrator type="path"/>''')

    integrator_nlos = fdent(f'''\
        <integrator type="transientpath">
            <integer name="block_size" value="1"/>
            <integer name="max_depth" value="{v('integrator_max_depth')}"/>
            <integer name="filter_depth" value="{v('integrator_filter_depth')}"/>
            <boolean name="discard_direct_paths" value="{v('integrator_discard_direct_paths')}"/>
            <boolean name="nlos_laser_sampling" value="{v('integrator_nlos_laser_sampling')}"/>
            <boolean name="nlos_hidden_geometry_sampling" value="{v('integrator_nlos_hidden_geometry_sampling')}"/>
            <boolean name="nlos_hidden_geometry_sampling_do_mis" value="{v('integrator_nlos_hidden_geometry_sampling_do_mis')}"/>
            <boolean name="nlos_hidden_geometry_sampling_includes_relay_wall" value="{v('integrator_nlos_hidden_geometry_sampling_includes_relay_wall')}"/>
        </integrator>''')

    if 'polarized' in v('mitsuba_variant'):
        def add_stokes(s, itype):
            return fdent('''\
                <integrator type="{itype}">
                    {s}
                </integrator>''', itype=itype, s=s)
        integrator_steady = add_stokes(integrator_steady, 'stokes')
        integrator_nlos = add_stokes(integrator_nlos, 'transientstokes')

    # relay wall
    geometry_names = list(map(lambda g: g['name'], v('geometry')))
    assert len(geometry_names) == len(set(geometry_names)), \
        'One or more geometry names is duplicated'
    relay_wall_name = v('relay_wall')
    assert relay_wall_name in geometry_names, \
        f'Relay wall is set to {relay_wall_name}, but must be one of {geometry_names}'

    dummy_lights_and_geometry_steady = fdent(f'''\
        <!-- Colored spheres to mark the laser and sensor positions -->
        <shape type="sphere">
            <point name="center" x="{v('sensor_x')}" y="{v('sensor_y')}" z="{v('sensor_z')}"/>
            <float name="radius" value="0.05"/>
            <bsdf type="diffuse" id="red">
                <rgb name="reflectance" value="1.0, 0.0, 0.0"/>
            </bsdf>
        </shape>
        <shape type="sphere">
            <point name="center" x="{v('laser_x')}" y="{v('laser_y')}" z="{v('laser_z')}"/>
            <float name="radius" value="0.05"/>
            <bsdf type="diffuse" id="blue">
                <rgb name="reflectance" value="0.0, 0.0, 1.0"/>
            </bsdf>
        </shape>
        
        <!-- Illuminate all the scene -->
        <emitter type="point">
            <rgb name="intensity" value="1.0, 1.0, 1.0"/>
            <point name="position" x="30" y="0" z="30"/>
        </emitter>
        <emitter type="point">
            <rgb name="intensity" value="1.0, 1.0, 1.0"/>
            <point name="position" x="-30" y="0" z="30"/>
        </emitter>
        <emitter type="point">
            <rgb name="intensity" value="1.0, 1.0, 1.0"/>
            <point name="position" x="0" y="30" z="30"/>
        </emitter>
        <emitter type="point">
            <rgb name="intensity" value="1.0, 1.0, 1.0"/>
            <point name="position" x="0" y="-30" z="30"/>
        </emitter>''')

    sensors_steady = fdent(f'''\
        <!-- Sensor 0: back view -->
        <sensor type="perspective">
            <string name="fov_axis" value="smaller"/>
            <float name="near_clip" value="0.01"/>
            <float name="far_clip" value="1000"/>
            <float name="fov" value="30"/>
            <transform name="to_world">
                <lookat origin="0, 0, 5"
                        target="0, 0, 0"
                            up="0, 1, 0"/>
            </transform>
            <sampler type="independent">
                <integer name="sample_count" value="512"/>
                <integer name="seed" value="{random_seed}"/>
            </sampler>
            <film type="hdrfilm">
                <integer name="width" value="512"/>
                <integer name="height" value="512"/>
                <rfilter name="rfilter" type="gaussian"/>
                <boolean name="high_quality_edges" value="false"/>
            </film>
        </sensor>

        <!-- Sensor 1: side view -->
        <sensor type="perspective">
            <string name="fov_axis" value="smaller"/>
            <float name="near_clip" value="0.01"/>
            <float name="far_clip" value="1000"/>
            <float name="fov" value="45"/>
            <transform name="to_world">
                <lookat origin="5, 0, 1.5"
                        target="0, 0, 1.5"
                            up="0, 1, 0"/>
            </transform>
            <sampler type="independent">
                <integer name="sample_count" value="512"/>
            </sampler>
            <film type="hdrfilm">
                <integer name="width" value="512"/>
                <integer name="height" value="512"/>
                <rfilter name="rfilter" type="gaussian"/>
                <boolean name="high_quality_edges" value="false"/>
            </film>
        </sensor>''')

    confocal_capture = 'false'
    if v('scan_type') == 'confocal' or v('scan_type') == 'exhaustive':
        confocal_capture = 'true'
    elif v('scan_type') != 'single':
        raise AssertionError(
            'scan_type should be one of {single|confocal|exhaustive}')
    sensor_nlos = fdent(f'''\
        <sensor type="nloscapturemeter">
            <sampler type="independent">
                <integer name="sample_count" value="{v('sample_count')}"/>
                <integer name="seed" value="{random_seed}"/>
            </sampler>

            <emitter type="projector">
                <rgb name="irradiance" value="1.0, 1.0, 1.0"/>
                <float name="fov" value="{0.2 if v('integrator_nlos_laser_sampling') else 2}"/>
            </emitter>

            <boolean name="confocal" value="{confocal_capture}"/>
            <boolean name="account_first_and_last_bounces" value="{v('account_first_and_last_bounces')}"/>
            <point name="sensor_origin" x="{v('sensor_x')}" y="{v('sensor_y')}" z="{v('sensor_z')}"/>
            <point name="laser_origin" x="{v('laser_x')}" y="{v('laser_y')}" z="{v('laser_z')}"/>
            <point name="laser_lookat_pixel" x="$laser_lookat_x" y="$laser_lookat_y" z="0"/>
            <film type="streakhdrfilm" name="streakfilm">
                <integer name="width" value="{v('sensor_width')}"/>
                <integer name="height" value="{v('sensor_height')}"/>
                <string name="pixel_format" value="rgb"/>
                <string name="component_format" value="float32"/>

                <integer name="num_bins" value="{v('num_bins')}"/>
                <boolean name="auto_detect_bins" value="{v('auto_detect_bins')}"/>
                <float name="bin_width_opl" value="{v('bin_width_opl')}"/>
                <float name="start_opl" value="{v('start_opl')}"/>

                <rfilter name="rfilter" type="box"/>
                <!-- NOTE: tfilters are not implemented yet -->
                <!-- <rfilter name="tfilter" type="box"/>  -->
                <boolean name="high_quality_edges" value="false"/>
            </film>
        </sensor>''')

    materials = get_materials()
    shapes_steady = []
    shapes_nlos = []
    for gdict in v('geometry'):
        def g(key):
            return s(gdict.get(key, None))

        name = g('name')
        is_relay_wall = name == relay_wall_name
        if is_relay_wall and g('mesh')['type'] != 'rectangle' and not quiet:
            print('WARNING: Relay wall does not work well with meshes that are '
                  'not of type "rectangle" because of wrong UV mapping. '
                  'Please make sure that you know what you are doing')
        shape_name = f'<!-- {name}{" (RELAY WALL)" if is_relay_wall else ""} -->'
        description = g('description')
        if description is not None and len(description.strip()) > 0:
            shape_name = fdent('''\
                {shape_name}
                <!--
                    {description}
                -->''', shape_name=shape_name, description=description.strip())

        material_id = g('material')['id']
        assert material_id in materials.keys(), \
            f'Geometry {name} has material {material_id}, but it must be one of {", ".join(materials.keys())}'
        shape_material = materials[material_id]
        material_keys = get_material_keys(shape_material)
        for key in material_keys:
            assert key in g('material'), \
                f'Material {material_id} for geometry {name} must have a value for {key}'
            shape_material.replace(f'${key}', g('material')[key])

        shape_transform = fdent(f'''\
            <transform name="to_world">
                <scale x="{g('scale') or 1.0}" y="{g('scale') or 1.0}" z="{g('scale') or 1.0}"/>
                <rotate x="1" angle="{g('rot_degrees_x') or 0.0}"/>
                <rotate y="1" angle="{g('rot_degrees_y') or 0.0}"/>
                <rotate z="1" angle="{g('rot_degrees_z') or 0.0}"/>
                <translate x="{g('displacement_x') or 0.0}" y="{g('displacement_y') or 0.0}" z="{g('displacement_z') or 0.0}"/>
            </transform>''')

        shape_contents_steady = f'{shape_material}\n{shape_transform}'
        newline = '\n'
        shape_contents_nlos = f'{shape_material}\n{shape_transform}{f"{newline}{sensor_nlos}" if is_relay_wall else ""}'

        if g('mesh')['type'] == 'obj':
            assert 'filename' in g('mesh'), \
                f'Missing mesh filename for geometry "{name}". ' \
                f'It is required because its mesh type is set as OBJ.'
            filename = g('mesh')['filename']
            assert os.path.isfile(filename), \
                f'{filename} does not exist for geometry "{name}"'

            def shapify(content, filename):
                return fdent('''\
                {shape_name}
                <shape type="obj">
                    <string name="filename" value="{filename}"/>

                    {content}
                </shape>''', shape_name=shape_name, filename=filename, content=content)

            shapes_steady.append(shapify(shape_contents_steady, filename))
            shapes_nlos.append(shapify(shape_contents_nlos, filename))
        elif g('mesh')['type'] == 'rectangle' or g('mesh')['type'] == 'sphere':
            def shapify(content):
                return fdent('''\
                {shape_name}
                <shape type="{shape_type}">
                    {content}
                </shape>''', shape_name=shape_name, content=content,
                             shape_type=g('mesh')['type'])

            shapes_steady.append(shapify(shape_contents_steady))
            shapes_nlos.append(shapify(shape_contents_nlos))
        else:
            raise AssertionError(
                f'Shape not yet supported: {g("mesh")["type"]}')

    shapes_steady = '\n\n'.join(shapes_steady)
    shapes_nlos = '\n\n'.join(shapes_nlos)

    file_steady = fdent('''\
        <!-- Auto-generated using TAL v{version} -->
        <scene version="2.2.1">
            {integrator_steady}

            {dummy_lights_and_geometry_steady}

            {shapes_steady}

            {sensors_steady}
        </scene>''',
                        version=tal.__version__,
                        integrator_steady=integrator_steady,
                        dummy_lights_and_geometry_steady=dummy_lights_and_geometry_steady,
                        shapes_steady=shapes_steady,
                        sensors_steady=sensors_steady)

    file_nlos = fdent('''\
        <!-- Auto-generated using TAL v{version} -->
        <scene version="2.2.1">
            {integrator_nlos}

            {shapes_nlos}
        </scene>''',
                      version=tal.__version__,
                      integrator_nlos=integrator_nlos,
                      shapes_nlos=shapes_nlos)

    return file_steady, file_nlos


def render_nlos_scene(config_path, args):
    config_path = os.path.abspath(config_path)

    assert os.path.exists(config_path), \
        f'{config_path} does not exist'

    path, name = os.path.split(config_path)
    candidate_filename = os.path.join(path, name, f'{name}.yaml')
    if os.path.isdir(config_path) and os.path.isfile(candidate_filename):
        config_path = candidate_filename

    assert os.path.isfile(config_path), \
        f'{config_path} is not a TAL config file'

    config_dir, config_filename = os.path.split(config_path)

    try:
        scene_config = yaml.safe_load(
            open(config_path, 'r')) or dict()
        scene_defaults = yaml.safe_load(
            open(local_file_path('render/scene_defaults.yaml'), 'r'))
    except yaml.YAMLError as exc:
        raise AssertionError(
            f'Invalid YAML format in TAL config file: {exc}') from exc
    scene_config = {**scene_defaults, **scene_config}

    if not args.dry_run:
        mitsuba_set_variant(scene_config['mitsuba_variant'])
    steady_xml, nlos_xml = get_scene_xml(
        scene_config, random_seed=args.seed, quiet=args.quiet)

    try:
        root_dir = os.path.join(
            config_dir, datetime.datetime.now().strftime(r'%Y%m%d-%H%M%S'))
        partial_results_dir = os.path.join(root_dir, 'partial')
        steady_dir = os.path.join(root_dir, 'steady')
        log_dir = os.path.join(root_dir, 'logs')
        os.mkdir(root_dir)
        os.mkdir(partial_results_dir)
        os.mkdir(steady_dir)
        os.mkdir(log_dir)
        shutil.copy(
            config_path,
            os.path.join(root_dir, f'{config_filename}.old'))
    except OSError as exc:
        raise AssertionError(f'Invalid permissions: {exc}') from exc

    try:
        steady_scene_xml = os.path.join(root_dir, 'steady_scene.xml')
        with open(steady_scene_xml, 'w') as f:
            f.write(steady_xml)

        nlos_scene_xml = os.path.join(root_dir, 'nlos_scene.xml')
        with open(nlos_scene_xml, 'w') as f:
            f.write(nlos_xml)

        laser_lookats = []
        name = scene_config['name']
        scan_type = scene_config['scan_type']
        num_bins = scene_config['num_bins']
        sensor_width = scene_config['sensor_width']
        sensor_height = scene_config['sensor_height']
        laser_width = scene_config['laser_width']
        laser_height = scene_config['laser_height']
        if scan_type == 'single':
            laser_width = 1
            laser_height = 1

        relay_wall = next(filter(
            lambda g: g['name'] == scene_config['relay_wall'],
            scene_config['geometry']))
        assert 'displacement_x' not in relay_wall and \
            'displacement_y' not in relay_wall and \
            'displacement_z' not in relay_wall and \
            'rot_degrees_x' not in relay_wall and \
            'rot_degrees_y' not in relay_wall and \
            'rot_degrees_z' not in relay_wall, \
            'Relay wall displacement/rotation is NYI'

        def get_grid_xyz(nx, ny, relay_wall_scale):
            p = relay_wall_scale
            xg = np.stack((np.linspace(-p, p, num=2*nx + 1)[1::2],)*ny, axis=1)
            yg = np.stack((np.linspace(-p, p, num=2*ny + 1)[1::2],)*nx, axis=0)
            assert xg.shape[0] == yg.shape[0] == nx and xg.shape[1] == yg.shape[1] == ny, \
                'Incorrect shapes'
            return np.stack([xg, yg, np.zeros((nx, ny))], axis=-1).astype(np.float32)

        def expand(vec, x, y):
            assert len(vec) == 3
            return vec.reshape(1, 1, 3).repeat(x, axis=0).repeat(y, axis=1)

        # FIXME(diego): rotate + translate (asssumes no rot/trans)
        # or use a more generalist approach that does not need to be rectangular
        sensor_grid_xyz = get_grid_xyz(
            sensor_width, sensor_height, relay_wall['scale'])
        laser_grid_xyz = get_grid_xyz(
            laser_width, laser_height, relay_wall['scale'])
        # FIXME(diego): rotate [0, 0, 1] by rot_degrees_x (assmes RW is a plane)
        # or use a more generalist approach
        sensor_grid_normals = expand(
            np.array([0, 0, 1]), sensor_width, sensor_height)
        laser_grid_normals = expand(
            np.array([0, 0, 1]), laser_width, laser_height)

        if scan_type == 'single':
            laser_lookat_x = \
                scene_config['laser_lookat_x'] or (sensor_width - 1) / 2
            laser_lookat_y = \
                scene_config['laser_lookat_y'] or (sensor_height - 1) / 2
            laser_lookats.append((laser_lookat_x, laser_lookat_y))
        elif scan_type == 'exhaustive' or scan_type == 'confocal':
            assert not (scan_type == 'confocal' and
                        (laser_width != sensor_width or
                            laser_height != sensor_height)), \
                'If using scan_type=confocal, sensor_{width|height} must match laser_{width|height}'
            for x in range(laser_width):
                for y in range(laser_height):
                    laser_lookats.append((x + 0.5, y + 0.5))
        else:
            raise AssertionError(
                'Invalid scan_type, must be one of {single|exhaustive|confocal}')

        experiment_name = scene_config['name']

        if args.do_steady_renders:
            def render_steady(render_name, sensor_index):
                if not args.quiet:
                    print(f'{render_name} for {experiment_name} steady render...')
                exr_path = os.path.join(partial_results_dir,
                                        f'{experiment_name}_{render_name}.exr')
                png_path = os.path.join(steady_dir,
                                        f'{experiment_name}_{render_name}.png')
                logfile = None
                if args.do_logging and not args.dry_run:
                    logfile = open(os.path.join(
                        log_dir, f'{experiment_name}_{render_name}.log'), 'w')
                run_mitsuba(steady_scene_xml, exr_path, dict(),
                            render_name, logfile, args, sensor_index)
                if args.do_logging and not args.dry_run:
                    logfile.close()
                if not args.dry_run:
                    image = read_mitsuba_bitmap(exr_path)
                    image = tonemap_ldr(image)
                    write_img(png_path, image)

            render_steady('back_view', 0)
            render_steady('front_view', 1)

        def partial_laser_dir(lx, ly):
            return os.path.join(partial_results_dir, f'{experiment_name}_L[{lx}][{ly}]'.replace('.', '_'))

        for i, (laser_lookat_x, laser_lookat_y) in tqdm(
                enumerate(laser_lookats), desc=f'Rendering {experiment_name} ({scan_type})...',
                ascii=True, total=len(laser_lookats)):
            try:
                exr_dir = partial_laser_dir(laser_lookat_x, laser_lookat_y)
                os.mkdir(exr_dir)
            except OSError as exc:
                raise AssertionError(f'Invalid permissions: {exc}') from exc
            defines = {
                'laser_lookat_x': laser_lookat_x,
                'laser_lookat_y': laser_lookat_y,
            }
            logfile = None
            if args.do_logging and not args.dry_run:
                logfile = open(os.path.join(
                    log_dir,
                    f'{experiment_name}_L[{laser_lookat_x}][{laser_lookat_y}].log'), 'w')
            run_mitsuba(nlos_scene_xml, exr_dir, defines,
                        f'Laser {i + 1} of {len(laser_lookats)}', logfile, args)
            if args.do_logging and not args.dry_run:
                logfile.close()

        if args.dry_run:
            return

        if not args.quiet:
            print('Merging partial results...')

        capture_data = NLOSCaptureData()
        capture_data.sensor_xyz = np.array([
            scene_config['sensor_x'],
            scene_config['sensor_y'],
            scene_config['sensor_z'],
        ], dtype=np.float32)
        capture_data.sensor_grid_xyz = sensor_grid_xyz
        capture_data.sensor_grid_normals = sensor_grid_normals
        capture_data.sensor_grid_format = GridFormat.X_Y_3
        capture_data.laser_xyz = np.array([
            scene_config['laser_x'],
            scene_config['laser_y'],
            scene_config['laser_z'],
        ], dtype=np.float32)
        capture_data.laser_grid_xyz = laser_grid_xyz
        capture_data.laser_grid_normals = laser_grid_normals
        capture_data.laser_grid_format = GridFormat.X_Y_3
        # NOTE(diego): we do not store volume information for now
        # capture_data.volume_format = VolumeFormat.X_Y_Z_3
        capture_data.delta_t = scene_config['bin_width_opl']
        capture_data.t_start = scene_config['start_opl']
        capture_data.t_accounts_first_and_last_bounces = \
            scene_config['account_first_and_last_bounces']
        capture_data.scene_info = {
            'tal_version': tal_version,
            'config': scene_config,
            'args': vars(args),
        }
        if scan_type == 'single':
            capture_data.H = read_mitsuba_streakbitmap(
                partial_laser_dir(*laser_lookats[0]))
            capture_data.H_format = HFormat.T_Sx_Sy
        elif scan_type == 'exhaustive' or scan_type == 'confocal':
            if scan_type == 'exhaustive':
                capture_data.H = np.empty(
                    (num_bins, laser_width, laser_height,
                     sensor_width, sensor_height),
                    dtype=np.float32)
                capture_data.H_format = HFormat.T_Lx_Ly_Sx_Sy
            elif scan_type == 'confocal':
                capture_data.H = np.empty(
                    (num_bins, laser_width, laser_height),
                    dtype=np.float32)
                capture_data.H_format = HFormat.T_Sx_Sy
            else:
                raise AssertionError

            for x in range(laser_width):
                for y in range(laser_height):
                    capture_data.H[:, x, y, ...] = np.squeeze(
                        read_mitsuba_streakbitmap(
                            partial_laser_dir(x + 0.5, y + 0.5)))
        else:
            raise AssertionError(
                'Invalid scan_type, must be one of {single|exhaustive|confocal}')

        hdf5_path = os.path.join(root_dir, f'{experiment_name}.hdf5')
        tal.io.write_capture(capture_data, hdf5_path,
                             file_format=FileFormat.HDF5_TAL)

        if not args.quiet:
            print(f'Stored result in {hdf5_path}')

        if args.keep_partial_results:
            return

        if not args.quiet:
            print(f'Cleaning partial results in {partial_results_dir}...')

        shutil.rmtree(partial_results_dir)

        if not args.quiet:
            print(f'All clean.')
    except KeyboardInterrupt:
        delete = None
        while delete is None:
            try:
                answer = input(f'Render cancelled. '
                               f'Delete the directory {root_dir}? (y/n): ')
                if answer.lower() == 'y':
                    delete = True
                elif answer.lower() == 'n':
                    delete = False
            except KeyboardInterrupt:
                pass
        if delete:
            shutil.rmtree(root_dir)



import tal
import os


def test(filename):
    print(f'Results for {filename}:')
    capture = tal.io.read_capture(
        os.path.join(os.path.split(__file__)[0], filename))
    print(f'H: {capture.H.shape}')
    print(f'H_format: {capture.H_format}')
    print(f'sensor_xyz: {capture.sensor_xyz}')
    print(f'sensor_grid_xyz: {capture.sensor_grid_xyz.shape}')
    print(f'sensor_grid_normals: {capture.sensor_grid_normals.shape}')
    print(f'sensor_grid_format: {capture.sensor_grid_format}')
    print(f'laser_xyz: {capture.laser_xyz}')
    print(f'laser_grid_xyz: {capture.laser_grid_xyz.shape}')
    print(f'laser_grid_normals: {capture.laser_grid_normals.shape}')
    print(f'laser_grid_format: {capture.laser_grid_format}')
    print(f'volume_format: {capture.volume_format}')
    # NOTE: idk why but delta_t upgrades to float64 here - its still stored as float32
    print(f'delta_t: {capture.delta_t}')
    print(f't_start: {capture.t_start}')
    print(
        f't_accounts_first_last_bounces: {capture.t_accounts_first_last_bounces}')
    print(f'scene_info: {capture.scene_info}')
    print()


test('data/NLOSDIRAC_nlosletters_nooffset.hdf5')
test('data/ZNLOS_Zdiffuse_withoffset.hdf5')
test('data/ZNLOS_Z_05_Exhaustive.hdf5')
test('data/ZNLOS_Z_05_Single.hdf5')
test('data/ZNLOS_Z_10_Confocal.hdf5')
test('data/ZNLOS_Z_10_Single.hdf5')



from enum import Enum


class FileFormat(Enum):
    AUTODETECT = 0
    HDF5_ZNLOS = 1
    HDF5_NLOS_DIRAC = 2
    HDF5_TAL = 3
    MAT_PHASOR_FIELDS = 4


class HFormat(Enum):
    UNKNOWN = 0
    T_Sx_Sy = 1  # confocal or not
    T_Lx_Ly_Sx_Sy = 2
    T_Si = 3  # confocal or not
    T_Si_Li = 4

    def time_dim(self) -> int:
        assert self in [HFormat.T_Sx_Sy,
                        HFormat.T_Lx_Ly_Sx_Sy,
                        HFormat.T_Si,
                        HFormat.T_Si_Li], \
            f'Unexpected HFormat {self}'
        return 0


class GridFormat(Enum):
    UNKNOWN = 0
    N_3 = 1
    X_Y_3 = 2


class VolumeFormat(Enum):
    UNKNOWN = 0
    N_3 = 1
    X_Y_Z_3 = 2
    X_Y_3 = 3

    def xyz_dim_is_last(self) -> bool:
        assert self in [VolumeFormat.N_3,
                        VolumeFormat.X_Y_Z_3,
                        VolumeFormat.X_Y_3]
        return True


class CameraSystem(Enum):
    STEADY = 0  # focused light
    # PHOTO_CAMERA = 1  # NYI (single-freq imaging)
    TRANSIENT = 2  # pulsed point light
    CONFOCAL_TIME_GATED = 3  # pulsed focused light

    def bp_accounts_for_d_2(self) -> bool:
        return self in [CameraSystem.STEADY, CameraSystem.CONFOCAL_TIME_GATED]

    def is_transient(self) -> bool:
        return self in [CameraSystem.TRANSIENT, CameraSystem.CONFOCAL_TIME_GATED]



def _get_setpath_location():
    from tal.config import ask_for_config, Config
    import os
    force_ask = False
    setpath_ok = False
    while not setpath_ok:
        setpath_location = os.path.join(
            ask_for_config(Config.MITSUBA2_TRANSIENT_NLOS_FOLDER,
                           force_ask=force_ask),
            'setpath.sh')
        if os.path.isfile(setpath_location):
            setpath_ok = True
        else:
            force_ask = True
            print(f'setpath.sh cannot be found in {setpath_location}.')
            print()
    return setpath_location


try:
    import mitsuba  # pyright: reportMissingImports=false
except ModuleNotFoundError:
    import sys
    import subprocess
    setpath_location = _get_setpath_location()

    command = ['env', '-i', '/bin/bash',
               '-c', f'source {setpath_location} && printenv']
    p = subprocess.check_output(command).decode('utf-8')
    for line in p.split('\n')[:-1]:
        (key, _, value) = line.partition('=')
        if key == 'PYTHONPATH':
            for directory in value.split(':')[:-1]:
                sys.path.append(directory)


def mitsuba_set_variant(s):
    import mitsuba
    mitsuba.set_variant(s)


def get_material_keys(s):
    import re
    return list(map(lambda e: e[1:],
                    re.compile(r'\$[a-zA-Z_]*').findall(s)))


def get_materials():
    from tal.util import fdent
    return {
        'white': fdent(f'''\
            <bsdf type="diffuse">
                <rgb name="reflectance" value="1.0, 1.0, 1.0"/>
            </bsdf>'''),
        'copper': fdent(f'''\
            <bsdf type="roughconductor">
                <string name="material" value="Cu"/>
                <string name="distribution" value="beckmann"/>
                <float name="alpha" value="$alpha"/>
            </bsdf>'''),
        'custom': '$text'
    }


def run_mitsuba(scene_xml_path, exr_path, defines,
                experiment_name, logfile, args, sensor_index=0):
    import re
    import time
    import subprocess
    import os
    from tqdm import tqdm
    # execute mitsuba command (sourcing setpath.sh before)
    num_threads = args.threads
    command = ['mitsuba',
               '-o', exr_path,
               '-s', str(sensor_index),
               '-t', str(num_threads)]
    for key, value in defines.items():
        command += ['-D', f'{key}={value}']
    command += [scene_xml_path]

    nice = args.nice
    command = ['nice', '-n', str(nice), " ".join(command)]

    setpath_location = _get_setpath_location()

    if args.dry_run:
        # add extra commas to make it easier to copy-paste
        command = ['/bin/bash', '-c',
                   f'"source \\"{setpath_location}\\" && {" ".join(command)}"']
        print(' '.join(command))
        return
    else:
        command = ['/bin/bash', '-c',
                   f'source "{setpath_location}" && {" ".join(command)}']

    if args.quiet:
        # simplified version, block until done rendering
        mitsuba_process = subprocess.Popen(
            command, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    else:
        # need to pass the command through stdbuf to be able to read the progress bar
        command = ['stdbuf', '-o0'] + command
        mitsuba_process = subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

        # read the progress bar and pass the info to the user through a tqdm bar
        # this is totally not overengineering-trust me-this saves so much time
        progress_re = re.compile(
            r'Rendering \[(=* *)\] \([\d\.]+\w+, ETA: ([\d\.]+\w+)\)')
        read_opl = defines.get('auto_detect_bins', False)
        if read_opl:
            opl_output = ''
            opl_re = re.compile(
                r'limits: \[(\d+\.\d+), \d+\.\d+\] with bin width (\d+\.\d+)')
        with tqdm(desc=experiment_name, total=100, ascii=True, leave=False,
                  bar_format='{desc} |{bar}| [{n:.2f}%{postfix}] ') as pbar:
            output = None
            while output is None or len(output) > 0:
                output = mitsuba_process.stdout.read(160)
                try:
                    output = output.decode('utf-8')
                except UnicodeDecodeError:
                    continue
                if logfile is not None:
                    logfile.write(output)
                    logfile.flush()
                if read_opl:
                    opl_output += output
                    matches = opl_re.findall(opl_output)
                    if len(matches) > 0:
                        start_opl, bin_width_opl = matches[-1]
                        start_opl = float(start_opl)
                        bin_width_opl = float(bin_width_opl)
                        print('Auto-detected histogram: '
                              f'start_opl={start_opl:.4f}, bin_width_opl={bin_width_opl:.6f}')
                        defines.update(start_opl=start_opl)
                        defines.update(bin_width_opl=bin_width_opl)
                        read_opl = False
                        del opl_output
                matches = progress_re.findall(output)
                if len(matches) > 0:
                    progress, eta = matches[-1]
                    completed = progress.count('=')
                    not_completed = progress.count(' ')
                    progress = 100 * completed / (completed + not_completed)
                    pbar.update(progress - pbar.n)
                    pbar.set_postfix_str(f'ETA: {eta}')
                    if not_completed == 0:
                        break
                    time.sleep(1)

    # wait for the process to end
    mitsuba_process.communicate()
    assert mitsuba_process.returncode == 0, \
        f'Mitsuba returned with error code {mitsuba_process.returncode}'
    if os.path.isdir(exr_path):
        # assume transient render
        n_exr_frames = len(os.listdir(exr_path))
        assert n_exr_frames > 0, 'No frames were rendered?'
    else:
        # assume steady state render
        assert os.path.isfile(exr_path), 'No frames were rendered?'


def read_mitsuba_bitmap(path: str):
    from mitsuba.core import Bitmap
    import numpy as np
    return np.array(Bitmap(path), copy=False)


def read_mitsuba_streakbitmap(path: str, exr_format='RGB'):
    """
    Reads all the images x-t that compose the streak image.

    :param dir: path where the images x-t are stored
    :return: a streak image of shape [time, width, height]
    """
    import re
    import glob
    import os
    import numpy as np
    from tqdm import tqdm

    # FIXME(diego): for now this assumes that the EXR that it reads
    # are in RGB format, and returns an image with 3 channels,
    # in the case of polarized light it should return something else
    if exr_format != 'RGB':
        raise NotImplementedError(
            'Formats different from RGB are not implemented')

    xtframes = glob.glob(os.path.join(
        glob.escape(path), f'frame_*.exr'))
    xtframes = sorted(xtframes,
                      key=lambda x: int(re.compile(r'\d+').findall(x)[-1]))
    number_of_xtframes = len(xtframes)
    first_img = read_mitsuba_bitmap(xtframes[0])
    streak_img = np.empty(
        (number_of_xtframes, *first_img.shape), dtype=first_img.dtype)
    with tqdm(desc=f'Reading {path}', total=number_of_xtframes, ascii=True) as pbar:
        for i_xtframe in range(number_of_xtframes):
            other = read_mitsuba_bitmap(xtframes[i_xtframe])
            streak_img[i_xtframe] = np.nan_to_num(other, nan=0.)
            pbar.update(1)

    # for now streak_img has dimensions (y, x, time, channels)
    assert streak_img.shape[-1] == 3, \
        f'Careful, streak_img has shape {streak_img.shape} (i.e. its probably not RGB as we assume, last dimension should be 3)'
    # and we want it as (time, x, y)
    return np.sum(np.transpose(streak_img), axis=0)



from tal.io.capture_data import NLOSCaptureData
from tal.enums import HFormat
import numpy as np


def filter_H_impl(data, filter_name, data_format, border, plot_filter, return_filter, **kwargs):
    if isinstance(data, NLOSCaptureData):
        assert data.H_format != HFormat.UNKNOWN or data_format is not None, \
            'H format must be specified through the NLOSCaptureData object or the data_format argument'
        H_format = (data_format
                    if data_format != HFormat.UNKNOWN
                    else None) or data.H_format
        H = data.H
        delta_t = data.delta_t
    else:
        assert data_format is not None and data_format != HFormat.UNKNOWN, \
            'If data is not a NLOSCaptureData object, H format must be specified through the data_format argument'
        H_format = data_format
        H = data
        delta_t = None

    assert border in ['erase', 'zero', 'edge'], \
        'border must be one of "erase", "zero" or "edge"'

    nt = H.shape[H_format.time_dim()]
    nt_pad = nt

    if filter_name == 'pf':
        wl_mean = kwargs.get('wl_mean', None)
        assert wl_mean is not None, \
            'For the "pf" filter, wl_mean must be specified'
        wl_sigma = kwargs.get('wl_sigma', None)
        wl_sigma = wl_sigma or wl_mean / np.sqrt(2)
        delta_t = kwargs.get('delta_t', None) or delta_t
        assert delta_t is not None, \
            'For the "pf" filter, delta_t must be specified through an NLOSCaptureData object or the delta_t argument'

        t_6sigma = int(np.round(6 * wl_sigma / delta_t))  # used for padding
        if t_6sigma % 2 == 1:
            t_6sigma += 1  # its easier if padding is even
        if return_filter:
            nt_pad = t_6sigma
        else:
            nt_pad = nt + 2 * (t_6sigma - 1)
        t_max = delta_t * (nt_pad - 1)
        t = np.linspace(start=0, stop=t_max, num=nt_pad)

        #   vvv Gaussian envelope (x = t - t_max/2, mu = 0, sigma = wl_sigma)
        gaussian_envelope = np.exp(-((t - t_max / 2) / wl_sigma) ** 2 / 2)
        K = gaussian_envelope / np.sum(gaussian_envelope) * \
            np.exp(2j * np.pi * t / wl_mean)
        #   ^^^ Pulse inside the Gaussian envelope (complex exponential)

        # center at zero (not in freq. domain but fftshift works)
        K = np.fft.ifftshift(K)
    else:
        raise AssertionError(
            'Unknown filter_name. Check the documentation for available filters')

    if plot_filter:
        import matplotlib.pyplot as plt
        K_show = np.fft.fftshift(K)
        cut = (nt_pad - t_6sigma) // 2
        if cut > 0:
            K_show = K_show[cut:-cut]
        plt.plot(t[:len(K_show)] - t[len(K_show) // 2], np.real(K_show),
                 c='b')
        plt.plot(t[:len(K_show)] - t[len(K_show) // 2], np.imag(K_show),
                 c='b', linestyle='--')
        plt.plot(t[:len(K_show)] - t[len(K_show) // 2], np.abs(K_show),
                 c='r')
        plt.show()
    if return_filter:
        return K

    padding = (nt_pad - nt)
    assert padding % 2 == 0
    padding //= 2

    # pad with edge values
    if H_format.time_dim() == 0:
        mode = None
        if border == 'edge':
            mode = 'edge'
        if border == 'zero' or border == 'erase':
            mode = 'constant'
        if mode:
            H_pad = np.pad(H,
                           ((padding, padding),) +  # first dim (time)
                           ((0, 0),) * (H.ndim - 1),  # other dims
                           mode=mode)
        K_shape = (nt_pad,) + (1,) * (H.ndim - 1)
    else:
        raise AssertionError('Unknown H_format')

    H_fft = np.fft.fft(H_pad, axis=0)
    K_fft = np.fft.fft(K)
    H_fft *= K_fft.reshape(K_shape)
    del K_fft
    HoK = np.fft.ifft(H_fft, axis=0)
    del H_fft

    # undo padding
    if H_format.time_dim() == 0:
        HoK = HoK[padding:-padding, ...]
        if border == 'erase':
            HoK[:padding//2] = 0
            HoK[-padding//2:] = 0
        return HoK
    else:
        raise AssertionError('Unknown H_format')



from enum import Enum
import os
import h5py
import yaml
import numpy as np
import inspect
from typing import Union, get_type_hints
from nptyping import NDArray, Shape
from tal.io.format import convert_dict, detect_dict_format
from tal.enums import FileFormat, HFormat, GridFormat, VolumeFormat


class LazyDataset:
    def __init__(self, dataset):
        self.read = False
        self.dataset = dataset

    def __get__(self, instance, owner):
        if not self.read:
            self.dataset = self.dataset[()]
            self.read = True
        return self.dataset

    def __getattr__(self, name):
        return getattr(self.dataset, name)


def read_hdf5(filename: str) -> dict:
    raw_data = h5py.File(filename, 'r')

    def parse(key, value):
        if isinstance(value, h5py.Empty):
            value = None
        elif isinstance(value, h5py.Group):
            value = {k: parse(k, v) for k, v in value.items()}
        else:
            if isinstance(value, h5py.Dataset):
                value = value[()]
                # FIXME(diego): unused, need to figure out how
                # to call do stuff like value.size or
                # isinstance(value, bytes) without reading the
                # whole dataset
                # value = LazyDataset(value)
            if isinstance(value, np.ndarray) and value.size == 1:
                value = np.squeeze(value)[()]

        # parse enums and strings
        if key in get_type_hints(NLOSCaptureData):
            key_class = get_type_hints(NLOSCaptureData)[key]
            if inspect.isclass(key_class) and issubclass(key_class, Enum):
                if isinstance(value, h5py.Empty):
                    value = key_class(0)
                else:
                    value = key_class(value)
        elif isinstance(value, bytes):
            value = yaml.safe_load(value)
        return value

    raw_data = dict((key, parse(key, raw_data.get(key)))
                    for key in raw_data.keys())
    return raw_data


def write_hdf5(capture_data: dict, filename: str):
    file = h5py.File(filename, 'w')
    for key, value in capture_data.items():
        if value is None:
            file[key] = h5py.Empty(float)
        elif isinstance(value, dict):
            file[key] = yaml.dump(value)
        elif isinstance(value, Enum):
            dt = h5py.enum_dtype(dict((item.name, item.value)
                                 for item in value.__class__), basetype='i')
            ds = file.create_dataset(key, (1,), dtype=dt)
            ds[0] = value.value
        else:
            file[key] = value
    file.close()


class NLOSCaptureData:

    """
    Type aliases
    """
    Float = np.float32
    TensorTSxSy = NDArray[Shape['T, Sx, Sy'], Float]
    TensorTLxLySxSy = NDArray[Shape['T, Lx, Ly, Sx, Sy'], Float]
    HType = Union[TensorTSxSy, TensorTLxLySxSy]
    MatrixN3 = NDArray[Shape['*, 3'], Float]
    TensorXY3 = NDArray[Shape['X, Y, 3'], Float]
    LaserGridType = Union[MatrixN3, TensorXY3]
    SensorGridType = Union[MatrixN3, TensorXY3]
    TensorXYZ3 = NDArray[Shape['X, Y, Z, 3'], Float]
    VolumeXYZType = Union[MatrixN3, TensorXYZ3]
    Array3 = NDArray[Shape['3'], Float]

    """ Capture data start (ignore _start) """
    _start: None = None  # used in as_dict()
    """
    TODO(diego): add docs
    """
    H: HType = None
    H_format: HFormat = None
    sensor_xyz: Array3 = None
    sensor_grid_xyz: SensorGridType = None
    sensor_grid_normals: SensorGridType = None
    sensor_grid_format: GridFormat = None
    laser_xyz: Array3 = None
    laser_grid_xyz: LaserGridType = None
    laser_grid_normals: LaserGridType = None
    laser_grid_format: GridFormat = None
    # volume_xyz : VolumeXYZType = None  # implemented as a property, see below
    volume_format: VolumeFormat = None
    delta_t: Float = None
    t_start: Float = None
    t_accounts_first_and_last_bounces: bool = None
    scene_info: dict = None  # additional information
    """
    Implemented scene_info keys:
    - 'original_format': str (e.g. 'HDF5_ZNLOS')
    - 'config': dict (original scene_config.yaml when generated using TAL)
    - 'args': dict (original args passed when generated using TAL)
    - 'volume': dict
        - 'center': Array3 (center of volume)
        - 'rotation': Array3 (rotation of volume formatted as per Z-NLOS - probably unused for now)
        - 'size': Array3 (size of volume - length of each XYZ side)
        - 'xyz': MatrixN3 (if ever volume_format is VolumeFormat.N_3, points will be stored here)
    """
    """ Capture data end (ignore _end) """
    _end: None = None  # used in as_dict()

    def __get_dict_keys(self):
        variables = list(self.__class__.__dict__.keys())
        variables = variables[variables.index(
            '_start') + 1: variables.index('_end')]
        return variables

    def __init__(self, filename: str = None, file_format: FileFormat = FileFormat.AUTODETECT):
        if filename is None:
            return

        # Read raw data and check its current format
        assert os.path.isfile(filename), f'Does not exist: {filename}'
        raw_data = read_hdf5(filename)
        if file_format == FileFormat.AUTODETECT:
            file_format = detect_dict_format(raw_data)

        # Convert data to HDF5_TAL format
        if file_format != FileFormat.HDF5_TAL:
            raw_data = convert_dict(
                raw_data, format_to=FileFormat.HDF5_TAL)

        own_dict_keys = self.__get_dict_keys()
        for key, value in raw_data.items():
            if key not in own_dict_keys:
                raise AssertionError(f'raw_data contains unknown key: {key}')
            setattr(self, key, value)

    def is_confocal(self):
        if self.H_format == HFormat.T_Lx_Ly_Sx_Sy:
            return False
        elif self.H_format == HFormat.T_Sx_Sy:
            return self.laser_grid_xyz.shape == self.sensor_grid_xyz.shape
        else:
            raise AssertionError('Invalid H_format')

    def as_dict(self):
        dict_keys = self.__get_dict_keys()
        return dict((key, getattr(self, key)) for key in dict_keys)

    @property
    def volume_xyz(self):
        try:
            return self.scene_info['volume']['xyz']
        except KeyError:
            pass

        try:
            # TODO(diego): return (X, Y, Z, 3) tensor using
            # self.scene_info['volume']['center' | 'rotation' | 'size']
            raise NotImplementedError('volume_xyz is not implemented yet')
        except KeyError:
            return None



import setuptools
from tal import __version__ as tal_version

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r", encoding="utf-8") as fh:
    requirements = fh.read().splitlines()

setuptools.setup(
    name="y-tal",
    version=tal_version,
    author="Diego Royo & Pablo Luesia",
    author_email="droyo@unizar.es",
    description="(Your) Transient Auxiliary Library - Analysis and processing of time-resolved light transport",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/diegoroyo/tal",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
    package_data={
        'tal': [
            '.tal.conf.example',
            'render/scene_defaults.yaml',
        ],
    },
    entry_points={
        'console_scripts': [
            'tal=tal.__main__:main'
        ]
    },
    python_requires=">=3.7",
    install_requires=requirements
)



from tal.render import create, render


def create_nlos_scene(folder_name, args):
    create.create_nlos_scene(folder_name, args)


def render_nlos_scene(config_path, args):
    render.render_nlos_scene(config_path, args)



import tal
from tal.reconstruct.pf import solve
from tal.reconstruct import get_volume
from tal.plot import zxy_interactive

import matplotlib.pyplot as plt
import numpy as np

if __name__ == "__main__":
    print('Start read')
    # data_path = 'D:/Documentos/nlos_dataset/rotated_mesh_cube_256x256/rotation[0]/data.hdf5'
    data_path = 'D:/Documentos/nlos_dataset/goal1-bigmirror.hdf5'
    data = tal.io.read_capture(data_path)
    print('Finish read')

    tal.plot.txy_interactive(data, cmap = 'hot')

    # row_x = np.linspace(-0.99609375, 0.99609375, 256, dtype = float)
    # row_a = np.stack([row_x, np.full(256, -0.99609375)], axis = -1)
    # row_b = np.stack([row_x, np.full(256, 0.99609375)], axis = -1)
    # plane_xy = np.linspace(row_a, row_b, 256)
    # plane_a = np.stack([plane_xy[:,:,0], plane_xy[:,:,1], np.full((256,256), 0.5)], axis = -1)
    # plane_b = np.stack([plane_xy[:,:,0], plane_xy[:,:,1], np.full((256,256), 1.5)], axis = -1)
    # V = np.linspace(plane_a, plane_b, 256)

    ni, nj = (32, 32)
    x_bounds = (-1, 1)
    y_bounds = (-2, -4)
    z_bounds = (-2, -4)
    upper_left_point = np.array([x_bounds[0], y_bounds[1], z_bounds[0]])
    upper_right_point = np.array([x_bounds[1], y_bounds[1], z_bounds[0]])
    lower_left_point = np.array([x_bounds[0], y_bounds[0], z_bounds[1]])
    lower_right_point = np.array([x_bounds[1], y_bounds[0], z_bounds[1]])
    upper_row = np.linspace(upper_left_point, lower_left_point, ni)
    lower_row = np.linspace(upper_right_point, lower_right_point, nj)
    V = np.linspace(upper_row, lower_row, nj).reshape(-1,3)


    # reconstruction = solve(data, 6, 20, get_volume(), 1, 2)
    reconstruction = solve(data, 16, 4, V, False, 3, 16)

    plt.figure()
    plt.imshow(np.abs(reconstruction).reshape((ni,nj)), cmap= 'hot')
    # plt.figure()
    # plt.imshow(np.angle(reconstruction), cmap= 'bwr')
    plt.show()

    # plot_zxy_interactive(np.abs(reconstruction))
    # plot_zxy_interactive(np.angle(reconstruction), cmap = 'bwr')